/* -*- c++ -*- */
/*
 * Expresso, a C++ Array template class library
 * Copyright (C) 1998,2004 Oskar Enoksson (enok@lysator.liu.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#ifndef EXPRESSO_FIXED_HDR
#define EXPRESSO_FIXED_HDR

#include "debug.h"
#include "tmpltool.h"
#include "multiptr.h"
#include "assign.h"
#include "iterator.h"
#include "permutation.h"
#include <cassert>
namespace esso {

///////////////////////////////////////////////////////////////////////
// Some template magic for use internally by Array<N,Fix<T> > methods
  namespace FixInternal {

    template<class T0, int M, bool Method>
    class FixInfo;

    template<class T0, int M>
    class FixInfo<T0,M,true> {
    public:
      enum { rank = TmplTools::ArrayInfo<T0>::rank };

      template<int N>
      static inline void FillStride(Ix<N> &stride) {
        stride[M] = TmplTools::ArrayInfoIndex<T0,M>::stride;
	FixInfo<T0,M+1,(M+1<=rank && M+1<N) >::FillStride(stride);
      }
      template<int N>
      static inline void FillLen(Ix<N> &len) {
        len[M] = TmplTools::ArrayInfoIndex<T0,M>::len;
	FixInfo<T0,M+1,(M+1<rank)>::FillLen(len);
      }
      template<int N>
      static inline IXTYPE Stride(const Ix<N> &stride) {
	return TmplTools::ArrayInfoIndex<T0,M>::stride;
      }
      template<int N>
      static inline IXTYPE Offset(const Ix<N> &ix) {
        return TmplTools::ArrayInfoIndex<T0,M>::stride*ix[M]+
	       FixInfo<T0,M+1,(M+1<=rank && M+1<N) >::Offset(ix);
      }
    };
    template<class T0, int M>
    class FixInfo<T0,M,false> {
    public:
      template<int N>
      static inline void FillStride(Ix<N> &stride) {}

      template<int N>
      static inline void FillLen(Ix<N> &len) {}

      template<int N>
      static inline IXTYPE Stride(const Ix<N> &stride) {
	return stride[M];
      }
      template<int N>
      static inline IXTYPE Offset(const Ix<N> &l) { return 0; }
    };

    template<bool Method>
    class Assign {
    public:
      template<int N, class T0, class TypeS>
      static inline bool Do(Array<N,Fix<T0> >&dest, const Array<N,TypeS> &src) {
        return dest.Refer(src);
      }
    };

    template<>
    class Assign<false> {
    public:
      template<int N, class T0, class TypeS>
      static inline bool Do(Array<N,Fix<T0> >&dest, const Array<N,TypeS> &src) {
        return dest.Clone(src);
      }
    };
  }

  /// Multidimensional, efficient array class for large data sets
  /** Use this array for efficient numerical computations
   * or other applications where powerful and efficient arrays are
   * needed for large data sets.
   *
   * This specialization of Array<N,T> is often more efficient than
   * the default Array type.
   *
   * The difference between Fix and standard Array is:
   *\li A Fix Array cannot permute its' ranks at runtime, the storage 
   * order of ranks in memory is fixed.
   *\li A Fix Array is in general faster than a flex array, because it's
   * lowest dimension always has stride 1.
   *\li A Fix Array can have a fixed length of one or more of it's lowest
   * ranks, a standard Array cannot.
   *\li A standard Array can refer to any slice of a larger array, Fix Arrays
   * can (in general) only refer to slices of other Fix arrays, and then only
   * to the lower ranks.
   *
   * As an example of fixing lower index bounds at compile time,
   * Array<3,Fix<int[5]> > declares a three-dimensional array whose 
   * lowest dimension is always of length 5. 
   * Or even Array<3,Fix<int[5][10]> > where the two lowest ranges are
   * given as template arguments.
   *
   * There may be two reasons for doing this. First it increases type
   * safety, if the lowest dimension range is known then the compiler should
   * be told about it. Second it may increase the compiler's ability to
   * optimize some loops.
   *
   * To retrieve a fortran-compatible pointer to first element, use member
   * function Start.
   */
  template<int N, class T0>
  class Array<N,Fix<T0> > {
  public:
    friend class Array<N,Fix<typename TmplTools::FlipConst<T0>::T> >;
/////////////////////////////
// typedefs and enums
    typedef T0 FixType;
    typedef typename TmplTools::ArrayInfo<T0>::T T;
    enum { hasdata=1 };
    enum { fixrank=TmplTools::ArrayInfo<T0>::rank };
    enum { fixtotlen=TmplTools::ArrayInfo<T0>::totlen };
    enum { varrank=N-fixrank };
    typedef Fix<T0> Type;
    typedef Array<N,Iterator<Type> > IteratorType;
    typedef TmplTools::TAssert<(varrank>0)> check;

    ///@name constructor
    ///@{
    /** Default constructor. Initialize array with no data and index bounds
     * zero */
    inline Array(): data(), origo(0) ,l(0) ,u(0) {
      FixInternal::FixInfo<FixType,0,true>::FillStride(stride);
    }
    /** Constructor with given number of elements and lower index bound = 0 */
    inline Array(const Ix<varrank> &len0);
    /** Constructor with given number of elements and given lower index 
	bound */
    inline Array(const Ix<varrank> &len0, const Ix<N> &l0);
    /** Copy constructor
	Creates a reference to the same memory area as \a src */
    inline Array(const Array &src):
      data(src.data), origo(src.origo), l(src.l), u(src.u), stride(src.stride)
      {}
    /** Copy constructor from another array type object
     * If possible, create a reference to the same memory area as 
     * \a src. If this is not possible, create a new memory area
     * and copy \a src 's elements */
    template<class TypeS>
    inline Array(const Array<N,TypeS> &src) {
      FixInternal::FixInfo<FixType,0,true>::FillStride(stride);
      FixInternal::Assign
	<(Array<N,TypeS>::hasdata &&
	  (TmplTools::IsSame<T,
	   typename Array<N,TypeS>::T>::check ||
	   TmplTools::IsSame<typename TmplTools::UnConst<T>::T,
	   typename Array<N,TypeS>::T>::check))>
	::Do(*this,src);
    }      
    /** 1-dimensional array of given length and lower bounds =0 
     * (only compiles for Array<1,T>) */
    inline Array(IXTYPE i0):
      data(fixtotlen*i0), origo(data.Data()),l(0) {
      TmplTools::TAssert<(varrank==1)>();
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=i0);
      FixInternal::FixInfo<FixType,0,(0<=fixrank)>::FillStride(stride);
      FixInternal::FixInfo<FixType,0,(0<fixrank)>::FillLen(u);
      u[fixrank]=i0;
    }
    /** 2-dimensional array of given length and lower bounds =0 
     * (only compiles for Array<2,T>) */
    inline Array(IXTYPE i0, IXTYPE i1):
      data(fixtotlen*i0*i1), origo(data.Data()),l(0) {
      TmplTools::TAssert<(varrank==2)>();
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=i0 && 0<=i1);
      FixInternal::FixInfo<FixType,0,(0<=fixrank)>::FillStride(stride);
      FixInternal::FixInfo<FixType,0,(0<fixrank)>::FillLen(u);
      u[fixrank]=i0;
      u[fixrank+1]=i1;
      stride[fixrank+1]=stride[fixrank]*i0;
    }
    /** 3-dimensional array of given length and lower bounds =0 
     * (only compiles for Array<3,T>) */
    inline Array(IXTYPE i0, IXTYPE i1, IXTYPE i2):
      data(fixtotlen*i0*i1*i2), origo(data.Data()),l(0) {
      TmplTools::TAssert<(varrank==3)>();
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=i0 && 0<=i1 && 0<=i2);
      FixInternal::FixInfo<FixType,0,(0<=fixrank)>::FillStride(stride);
      FixInternal::FixInfo<FixType,0,(0<fixrank)>::FillLen(u);
      u[fixrank]=i0;
      u[fixrank+1]=i1;
      u[fixrank+2]=i2;
      stride[fixrank+1]=stride[fixrank]*i0;
      stride[fixrank+2]=stride[fixrank+1]*i1;
    }
    /** 4-dimensional array of given length and lower bounds =0 
     * (only compiles for Array<4,T>) */
    inline Array(IXTYPE i0, IXTYPE i1, IXTYPE i2, IXTYPE i3):
      data(fixtotlen*i0*i1*i2*i3), origo(data.Data()),l(0) {
      TmplTools::TAssert<(varrank==4)>();
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=i0 && 0<=i1 && 0<=i2 && 0<=i3);
      FixInternal::FixInfo<FixType,0,(0<=fixrank)>::FillStride(stride);
      FixInternal::FixInfo<FixType,0,(0<fixrank)>::FillLen(u);
      u[fixrank]=i0;
      u[fixrank+1]=i1;
      u[fixrank+2]=i2;
      u[fixrank+3]=i3;
      stride[fixrank+1]=stride[fixrank]*i0;
      stride[fixrank+2]=stride[fixrank+1]*i1;
      stride[fixrank+3]=stride[fixrank+2]*i2;
    }
    /** 5-dimensional array of given length and lower bounds =0 
     * (only compiles for Array<5,T>) */
    inline Array(IXTYPE i0, IXTYPE i1, IXTYPE i2, IXTYPE i3, IXTYPE i4):
      data(fixtotlen*i0*i1*i2*i3*i4), origo(data.Data()),l(0) {
      TmplTools::TAssert<(varrank==5)>();
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=i0 && 0<=i1 && 0<=i2 && 0<=i3 && 0<=i4);
      FixInternal::FixInfo<FixType,0,(0<=fixrank)>::FillStride(stride);
      FixInternal::FixInfo<FixType,0,(0<fixrank)>::FillLen(u);
      u[fixrank]=i0;
      u[fixrank+1]=i1;
      u[fixrank+2]=i2;
      u[fixrank+3]=i3;
      u[fixrank+4]=i4;
      stride[fixrank+1]=stride[fixrank]*i0;
      stride[fixrank+2]=stride[fixrank+1]*i1;
      stride[fixrank+3]=stride[fixrank+2]*i2;
      stride[fixrank+4]=stride[fixrank+3]*i3;
    }
    /** 6-dimensional array of given length and lower bounds =0 
     * (only compiles for Array<6,T>) */
    inline Array(IXTYPE i0, IXTYPE i1, IXTYPE i2, IXTYPE i3, IXTYPE i4, IXTYPE i5):
      data(fixtotlen*i0*i1*i2*i3*i4*i5), origo(data.Data()),l(0) {
      TmplTools::TAssert<(varrank==6)>();
      EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=i0 && 0<=i1 && 0<=i2 && 0<=i3 && 0<=i4 && 0<=i5);
      FixInternal::FixInfo<FixType,0,(0<=fixrank)>::FillStride(stride);
      FixInternal::FixInfo<FixType,0,(0<fixrank)>::FillLen(u);
      u[fixrank]=i0;
      u[fixrank+1]=i1;
      u[fixrank+2]=i2;
      u[fixrank+3]=i3;
      u[fixrank+4]=i4;
      u[fixrank+5]=i5;
      stride[fixrank+1]=stride[fixrank]*i0;
      stride[fixrank+2]=stride[fixrank+1]*i1;
      stride[fixrank+3]=stride[fixrank+2]*i2;
      stride[fixrank+4]=stride[fixrank+3]*i3;
      stride[fixrank+5]=stride[fixrank+4]*i4;
    }
    ///@}
    ///@name const
    ///@{
    /** Lower index bounds
     * L(n) is the lowest index allowed on dimension \a n
     */
    inline IXTYPE L(DIMT n) const { 
      EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N); 
      return l[n]; 
    }
    /** Upper index bounds
     * U(n) is one more than the largest index allowed on dimension \a n
     */
    inline IXTYPE U(DIMT n) const { 
      EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N); 
      return u[n]; 
    }
    /** Convenience function giving the number of array elements along
     * given dimension.
     * Simply put, Len(n) is U(n) - L(n)
     */
    inline IXTYPE Len(DIMT n) const { 
      EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N); 
      return u[n]-l[n]; 
    }
    /**\internal Number of elements between each position along dimension \a n
     * Stride(n) may be useful in some rare circumstances where someone
     * wishes to use pointer arithmetics on element data.
     */
    inline IXTYPE Stride(DIMT n) const { 
      EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N); 
      return stride[n]; 
    }
    /** Pointer to element at index (0,0,...,0)
     * Useful in some rare circumstances. Note that this pointer does
     * not point to a valid element if (0,0,...,0) is outside valid
     * index range. */
    inline T * Origo() const { return origo; }
    /** Pointer to first valid element i.e. element at lowest index bound 
     *
     * Use this function to get pointer to pass to fortran functions.
     */
    inline T * Start() const { return & operator [](l); }
    /**\internal 
     * Pointer to underlying, reference counting data object. 
     * Rarely useful */
    inline const MultiPtr<T> &Data() const { return data; }
    /**@return true if this array has no data allocated */
    inline bool IsNull() const { return data.Data()==0; }
    ///@}
    ///@name nonconst
    ///@{
    /** Release this object from it's element data and set index
     * bounds to zero 
     *@return always returns true */
    inline bool Dealloc();
    /** Release this object from it's element data and allocate a
     * new data set with given size (lower index bounds zero) 
     *@return always returns true */
    inline bool Realloc(const Ix<varrank> &len0);
    /** Release this object from it's element data and allocate a
     * new data set with given size and given lower index bounds
     *@return always returns true */
    inline bool Realloc(const Ix<varrank> &len0, const Ix<N> &l0);
    /**\internal
     * Release this object from it's element data and set it to
     * point to the given data with given size and lower index bounds 
     * (rarely useful)
     *@return always returns true */
    inline bool Realloc(const Ix<varrank> &len0, const Ix<N> &l0,
		       const MultiPtr<T> &data0, T * origo0);
    /** Release this object from it's element data and allocate a
     * new, uninitialized data set with index bounds identical to 
     * those of \a src.
     *@return true if successful or false if operation failed */
    template<class TypeS>
    inline bool Realloc(const Array<N,TypeS> &src);
    /** Reinitialize this object as an identical copy of \a src
     * (refering to the same element data as \a src) 
     *@return true if successful or false if operation failed */
    template<class TypeS>
    inline bool Refer(const Array<N,TypeS> &src);
    /** Same as Clone(*this).
     *@return always returns true */
    inline bool Clone();
    /** Release this object from it's element data and allocate a
     * new data set with index bounds identical to those of \a src.
     * Finally copy the element data from \a src
     *
     * Note that this is the same as Realloc followed by element-wise
     * copy.
     *@return true if successful or false if operation failed */
    template<class TypeS>
    inline bool Clone(const Array<N,TypeS> &src);
    /** Create an identical reference to \a src data if possible (same as 
     * Refer(src) ), otherwise create a new data area and copy \a src 's
     * elements into it (same as Clone(src)) */
    template<class TypeS>
    inline Array & operator =(const Array<N,TypeS> &src) {
      FixInternal::Assign
	<(Array<N,TypeS>::hasdata &&
	  (TmplTools::IsSame<T,typename Array<N,TypeS>::T>::check ||
	   TmplTools::IsSame<typename TmplTools::UnConst<T>::T,
	   typename Array<N,TypeS>::T>::check))>
	::Do(*this,src);
      return *this;
    }
    /** Create an identical reference to \a src data (same as Refer(src)) */
    inline Array & operator =(const Array &src) {
      data=src.data;
      origo=src.origo;
      l=src.l;
      u=src.u;
      stride=src.stride;
      return *this;
    }
    /** Shift data and valid index range along dimension \a n */
    inline Array & Shift(DIMT n, IXTYPE i);
    /** Shift data and valid index range along all dimensions */
    inline Array & Shift(const Ix<N> &ix);
    /** Restrict valid index range along dimension \a n .
     * Resulting index range along \a n is intersection of
     * \c [l0,u0[ and \c [L(n),U(n)[.
     */
    inline Array & Restrict(DIMT n, IXTYPE l0, IXTYPE u0);
    /** Restrict valid index range along all dimensions */
    inline Array & Restrict(const Ix<N> &l0, const Ix<N> &u0);
    /** Restrict lower index bound along dimension \a n */
    inline Array & RestrictL(DIMT n, IXTYPE l0);
    /** Restrict lower index bound along all dimensions */
    inline Array & RestrictL(const Ix<N> &l0);
    /** Restrict upper index bound along dimension \a n */
    inline Array & RestrictU(DIMT n, IXTYPE u0);
    /** Restrict upper index bound along all dimensions */
    inline Array & RestrictU(const Ix<N> &u0);
    ///@}    
    /** Get 1-dimensional array element at given position */
    inline T & operator ()(IXTYPE i0) const {
      TmplTools::TAssert<(N==1)>();
      EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[0]<=i0 && i0<u[0]);
      return *(origo+
	       i0);
    }
    /** Get 2-dimensional array element at given position */
    inline T & operator ()(IXTYPE i0, IXTYPE i1) const {
      TmplTools::TAssert<(N==2)>();
      EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[0]<=i0 && i0<u[0]
		   && l[1]<=i1 && i1<u[1]);
      return *(origo+
	       FixInternal::FixInfo<T0,1,(fixrank>=1)>::Stride(stride)*i1+
	       i0);
    }
    /** Get 3-dimensional array element at given position */
    inline T & operator ()(IXTYPE i0, IXTYPE i1, IXTYPE i2) const {
      TmplTools::TAssert<(N==3)>();
      EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[0]<=i0 && i0<u[0]
		   && l[1]<=i1 && i1<u[1]
		   && l[2]<=i2 && i2<u[2]);
      return *(origo+
	       FixInternal::FixInfo<T0,2,(fixrank>=2)>::Stride(stride)*i2+
	       FixInternal::FixInfo<T0,1,(fixrank>=1)>::Stride(stride)*i1+
	       i0);
    }
    /** Get 4-dimensional array element at given position */
    inline T & operator ()(IXTYPE i0, IXTYPE i1, IXTYPE i2, IXTYPE i3) const {
      TmplTools::TAssert<(N==4)>();
      EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[0]<=i0 && i0<u[0]
		   && l[1]<=i1 && i1<u[1]
		   && l[2]<=i2 && i2<u[2]
		   && l[3]<=i3 && i3<u[3]);
      return *(origo+
	       FixInternal::FixInfo<T0,3,(fixrank>=3)>::Stride(stride)*i3+
	       FixInternal::FixInfo<T0,2,(fixrank>=2)>::Stride(stride)*i2+
	       FixInternal::FixInfo<T0,1,(fixrank>=1)>::Stride(stride)*i1+
	       i0);
    }
    /** Get 5-dimensional array element at given position */
    inline T & operator ()(IXTYPE i0, IXTYPE i1, IXTYPE i2, IXTYPE i3, IXTYPE i4) const {
      TmplTools::TAssert<(N==5)>();
      EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[0]<=i0 && i0<u[0]
		   && l[1]<=i1 && i1<u[1]
		   && l[2]<=i2 && i2<u[2]
		   && l[3]<=i3 && i3<u[3]
		   && l[4]<=i4 && i4<u[4]);
      return *(origo+
	       FixInternal::FixInfo<T0,4,(fixrank>=4)>::Stride(stride)*i4+
	       FixInternal::FixInfo<T0,3,(fixrank>=3)>::Stride(stride)*i3+
	       FixInternal::FixInfo<T0,2,(fixrank>=2)>::Stride(stride)*i2+
	       FixInternal::FixInfo<T0,1,(fixrank>=1)>::Stride(stride)*i1+
	       i0);
    }
    /** Get 6-dimensional array element at given position */
    inline T & operator ()(IXTYPE i0, IXTYPE i1, IXTYPE i2, IXTYPE i3, IXTYPE i4, IXTYPE i5) const {
      TmplTools::TAssert<(N==6)>();
      EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[0]<=i0 && i0<u[0]
		   && l[1]<=i1 && i1<u[1]
		   && l[2]<=i2 && i2<u[2]
		   && l[3]<=i3 && i3<u[3]
		   && l[4]<=i4 && i4<u[4]
		   && l[5]<=i5 && i5<u[5]);
      return *(origo+
	       FixInternal::FixInfo<T0,5,(fixrank>=5)>::Stride(stride)*i5+
	       FixInternal::FixInfo<T0,4,(fixrank>=4)>::Stride(stride)*i4+
	       FixInternal::FixInfo<T0,3,(fixrank>=3)>::Stride(stride)*i3+
	       FixInternal::FixInfo<T0,2,(fixrank>=2)>::Stride(stride)*i2+
	       FixInternal::FixInfo<T0,1,(fixrank>=1)>::Stride(stride)*i1+
	       i0);
    }
    /** Get element at given position */
    inline T & operator [](const Ix<N> &ix) const {
#ifdef EXPRESSO_DEBUG_CHECK_INDEXING_RANGE
      for(int k=0; k<N; k++) EXPRESSO_DEBUG_CHECK_INDEXING_RANGE(l[k]<=ix[k] && ix[k]<u[k]);
#endif
      return *(origo+
	       IndexInternal::Recurse<fixrank+1,N>::SumProd(stride,ix)+
               FixInternal::FixInfo<FixType,0,true>::Offset(ix));
    }

/////////////////////////////
// Constructors

/////////////////////////////
// Assignment operator

/////////////////////////////
// Private data
  private:
    /** Reference counting data object */
    MultiPtr<T> data;
    /** Pointer to element at index (0,0,...,0) */
    T *origo;
    /** Lower index bound */
    Ix<N> l;
    /** Upper index bound */
    Ix<N> u;
    /** Stride, i.e. number of elements between each index along each
     * dimension. */
    Ix<N> stride;
  };

  template<int N, class T0>
  inline bool
  Array<N,Fix<T0> >::Dealloc() {
    data=MultiPtr<T>();
    origo=data.Data();
    l=0;
    u=0;
    for(int n=fixrank+1; n<N; n++)
      stride[n]=stride[fixrank];
    return true;
  }

  template<int N, class T0>
  inline bool
  Array<N,Fix<T0> >::Clone() {
    return Clone(*this);
  }

  template<int N, class T0>
  inline bool
  Array<N,Fix<T0> >::Realloc(const Ix<varrank> &len0) {
#ifdef EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE
    for(int k=0; k<varrank; k++) EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=len0[k]);
#endif
    data=MultiPtr<T>(fixtotlen*IndexInternal::Recurse<0,varrank>::Prod(len0));
    origo=data.Data();
    l=0;
    FixInternal::FixInfo<FixType,0,0<fixrank>::FillLen(u);
    for(int n=fixrank; n<N; n++)
      u[n]=len0[n-fixrank];
    for(int n=fixrank+1; n<N; n++)
      stride[n]=stride[n-1]*len0[n-1-fixrank];
    return true;
  }

  template<int N, class T0>
  inline bool
  Array<N,Fix<T0> >::Realloc(const Ix<varrank> &len0, const Ix<N> &l0) {
#ifdef EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE
    for(int k=0; k<varrank; k++) EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=len0[k]);
#endif
    Realloc(len0);
    l=l0;
    for(int n=0; n<N; n++)
      u[n]+=l0[n];
    origo=origo-IndexInternal::Recurse<fixrank+1,N>::SumProd(stride,l)
               -FixInternal::FixInfo<FixType,0,true>::Offset(l);
    return true;
  }

  template<int N, class T0>
  inline bool
  Array<N,Fix<T0> >::Realloc(const Ix<varrank> &len0, const Ix<N> &l0,
                            const MultiPtr<T> &data0, T *origo0) {
#ifdef EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE
    for(int k=0; k<varrank; k++) EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=len0[k]);
#endif
    data=data0;
    origo=origo0;
    l=l0;
    FixInternal::FixInfo<FixType,0,0<fixrank>::FillLen(u);
    for(int n=0; n<fixrank; n++)
      u[n]+=l[n];
    for(int n=fixrank; n<N; n++)
      u[n]=len0[n-fixrank]+l[n];
    for(int n=fixrank+1; n<N; n++)
      stride[n]=stride[n-1]*len0[n-1-fixrank];
    return true;
  }

  template<int N, class T0>
  template<class TypeS>
  inline bool
  Array<N,Fix<T0> >::Refer(const Array<N,TypeS> &src) {
    if(src.IsNull()) {
      Dealloc();
    } else {
      // Check so that fixed ranks have the correct strides
      for(int n=0; n<=fixrank; n++) {
        if(src.Stride(n)!=stride[n]) {
	  Dealloc();
	  return false;
        }
      }
      // Check so that variable ranks are in order
      for(int n=fixrank+1; n<N; n++) {
        if(src.Stride(n)<src.Stride(n-1) ||
	   src.Stride(n)%src.Stride(n-1) != 0) {
	  Dealloc();
	  return false;
	}
      }
      data=src.Data();
      origo=src.Origo();
      for(int n=0; n<N; n++) {
        l[n]=src.L(n);
        u[n]=src.U(n);
        stride[n]=src.Stride(n);
      }
    }
    return true;
  }

  template<int N, class T0>
  template<class TypeS>
  inline bool
  Array<N,Fix<T0> >::Realloc(const Array<N,TypeS> &src) {
    int totsize;
    for(int n=0; n<fixrank; n++)
      if(src.Len(n)!=0 && src.Len(n)*stride[n]!=stride[n+1]) {
	return false;
      }
    for(int n=0; n<N; n++) {
      l[n]=src.L(n);
      u[n]=src.U(n);
    }
    totsize=stride[fixrank]*Len(fixrank);
    for(int n=fixrank+1; n<N; n++) {
      stride[n]=totsize;
      totsize*=Len(n);
    }
    data=MultiPtr<T>(totsize);
    origo=data.Data()-IndexInternal::Recurse<fixrank+1,N>::SumProd(stride,l)-
                      FixInternal::FixInfo<FixType,0,true>::Offset(l);
    return true;
  }

  template<int N, class T0>
  template<class TypeS>
  inline bool
  Array<N,Fix<T0> >::Clone(const Array<N,TypeS> &src) {
    Array<N,Fix<T0> > foo;
    if(!foo.Realloc(src))
      return false;
    if(!foo.IsNull())
      foo<<=src;
    *this=foo;
    return true;
  }

  template<int N, class T0>
  inline Array<N,Fix<T0> > &
  Array<N,Fix<T0> >::Shift(DIMT n, IXTYPE i) {
    EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N);
    l[n]+=i;
    u[n]+=i;
    origo-=i*stride[n];
    return *this; 
  }

  template<int N, class T0>
  inline Array<N,Fix<T0> > &
  Array<N,Fix<T0> >::Shift(const Ix<N> &ix) {
    for(int n=0; n<N; n++)
      Shift(n,ix[n]);
    return *this;  
  } 

  template<int N, class T0>
  inline Array<N,Fix<T0> > &
  Array<N,Fix<T0> >::RestrictL(DIMT n, IXTYPE l0) {
    EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N);
    l[n]=std::max(l[n],l0);
    if(l[n]>=u[n]) {
      l[n]=u[n];
      data.clear();
    }
    return *this;
  }  

  template<int N, class T0>
  inline Array<N,Fix<T0> > &
  Array<N,Fix<T0> >::RestrictL(const Ix<N> &l0) {
    for(int n=0; n<N; n++) {
      l[n]=std::max(l[n],l0[n]);
      if(l[n]>=u[n]) {
	l[n]=u[n];
	data.clear();
	break;
      }
    }
    return *this;
  }

  template<int N, class T0>
  inline Array<N,Fix<T0> > &
  Array<N,Fix<T0> >::RestrictU(DIMT n, IXTYPE u0) {
    EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N);
    u[n]=std::min(u[n],u0);
    if(u[n]<=l[n]) {
      u[n]=l[n];
      data.clear();
    }
    return *this;
  }  

  template<int N, class T0>
  inline Array<N,Fix<T0> > &
  Array<N,Fix<T0> >::RestrictU(const Ix<N> &u0) {
    for(int n=0; n<N; n++) {
      u[n]=std::min(u[n],u0[n]);
      if(u[n]<=l[n]) {
	u[n]=l[n];
	data.clear();
	break;
      }
    }
    return *this;
  }

  template<int N, class T0>
  inline Array<N,Fix<T0> > &
  Array<N,Fix<T0> >::Restrict(DIMT n, IXTYPE l0, IXTYPE u0) {
    EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N);
    l[n]=std::max(l[n],l0);
    u[n]=std::min(u[n],u0);
    if(l[n]>=u[n]) {
      l[n]=u[n];
      data.clear();
    }
    return *this;
  }  

  template<int N, class T0>
  inline Array<N,Fix<T0> > &
  Array<N,Fix<T0> >::Restrict(const Ix<N> &l0, const Ix<N> &u0) {
    for(int n=0; n<N; n++) {
      l[n]=std::max(l[n],l0[n]);
      u[n]=std::min(u[n],u0[n]);
      if(l[n]>=u[n]) {
	l[n]=u[n];
	data.clear();
	break;
      }
    }
    return *this;
  }

  template<int N, class T0>
  inline 
  Array<N,Fix<T0> >::Array(const Ix<varrank> &len0):
    data(fixtotlen*IndexInternal::Recurse<0,varrank>::Prod(len0)), origo(data.Data()),l(0)
  {
#ifdef EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE
    for(int k=0; k<varrank; k++) EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=len0[k]);
#endif
    FixInternal::FixInfo<FixType,0,(0<=fixrank)>::FillStride(stride);
    FixInternal::FixInfo<FixType,0,(0<fixrank)>::FillLen(u);
    for(int n=fixrank; n<N; n++)
      u[n]=len0[n-fixrank];
    for(int n=fixrank+1; n<N; n++)
      stride[n]=stride[n-1]*len0[n-1-fixrank];
  }

  template<int N, class T0>
  inline 
  Array<N,Fix<T0> >::Array(const Ix<varrank> &len0, const Ix<N> &l0):
    data(fixtotlen*IndexInternal::Recurse<0,varrank>::Prod(len0)), origo(data.Data()),l(l0)
  {
#ifdef EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE
    for(int k=0; k<varrank; k++) EXPRESSO_DEBUG_CHECK_NONEGATIVE_SIZE(0<=len0[k]);
#endif
    FixInternal::FixInfo<FixType,0,(0<=fixrank)>::FillStride(stride);
    FixInternal::FixInfo<FixType,0,(0<fixrank)>::FillLen(u);
    for(int n=0; n<fixrank; n++)
      u[n]+=l[n];
    for(int n=fixrank; n<N; n++)
      u[n]=len0[n-fixrank]+l[n];
    for(int n=fixrank+1; n<N; n++)
      stride[n]=stride[n-1]*len0[n-1-fixrank];
    origo=origo-IndexInternal::Recurse<fixrank+1,N>::SumProd(stride,l)-
                FixInternal::FixInfo<FixType,0,true>::Offset(l);
  }

///////////////////////////////////////////////////////////////////////
// Functions on Fix arrays:

  template<int N, class T0>
  inline Array<N,typename TmplTools::ArrayInfo<T0>::T>
  Reverse(const Array<N,Fix<T0> > &src, DIMT n) {
    return Array<N,typename TmplTools::ArrayInfo<T0>::T>(src).Reverse(n);
  }

  template<int N, class T0>
  inline Array<N,typename TmplTools::ArrayInfo<T0>::T>
  Reverse(const Array<N,Fix<T0> > &src, const Ix<N,bool> &nx) {
    return Array<N,typename TmplTools::ArrayInfo<T0>::T>(src).Reverse(nx);
  }

  template<int N, class T0>
  inline Array<N,typename TmplTools::ArrayInfo<T0>::T>
  Flip(const Array<N,Fix<T0> > &src, DIMT n) {
    return Array<N,typename TmplTools::ArrayInfo<T0>::T>(src).Flip(n);
  }

  template<int N, class T0>
  inline Array<N,typename TmplTools::ArrayInfo<T0>::T>
  Flip(const Array<N,Fix<T0> > &src, const Ix<N,bool> &nx) {
    return Array<N,typename TmplTools::ArrayInfo<T0>::T>(src).Flip(nx);
  }

  template<int N, class T0>
  inline Array<N,typename TmplTools::ArrayInfo<T0>::T>
  Restride(const Array<N,Fix<T0> > &src, DIMT n, IXTYPE s) {
    return Array<N,typename TmplTools::ArrayInfo<T0>::T>
      (src).Restride(n,s);
  }

  template<int N, class T0>
  inline Array<N,typename TmplTools::ArrayInfo<T0>::T>
  Restride(const Array<N,Fix<T0> > &src, const Ix<N> &sx) {
    return Array<N,typename TmplTools::ArrayInfo<T0>::T>
      (src).Restride(sx);
  }

  template<int N, class T0>
  Array<N,typename TmplTools::ArrayInfo<T0>::T>
  Permute(const Array<N,Fix<T0> > &src, DIMT n1, DIMT n2) {
    return Array<N,typename TmplTools::ArrayInfo<T0>::T>(src).Permute(n1,n2);
  }

  template<int N, class T0>
  Array<N,typename TmplTools::ArrayInfo<T0>::T>
  Permute(const Array<N,Fix<T0> > &src, const Permutation<N> &nx) {
    return Array<N,typename TmplTools::ArrayInfo<T0>::T>(src).Permute(nx);
  }

#ifdef __KCC
  template<class T>
  class KAI_workaround;

  template<class T0>
  class KAI_workaround<Fix<T0> > {
  public:
    typedef typename TmplTools::ArrayInfo<T0>::T ReturnType;
    /*
    template<int N>
    static Array<N-1,Fix<T0> >
    Rmdim(const Array<N,Fix<T0> > &src, IXTYPE i) {
      Array<N-1,Fix<T0> > ret;
      Ix<N-1> len0,l0;
      for(int n=0; n<N-1; n++)
	len0[n]=src.Len(n);
      for(int n=0; n<N-1; n++)
	l0[n]=src.L(n);
      
      ret.Realloc(len0,l0,src.Data(),src.Origo());
      return ret;
    }
    */
    template<int N>
    static Array<N-1,ReturnType>
    Rmdim(const Array<N,Fix<T0> > &src, DIMT n, IXTYPE i) {
      return Array<N-1,ReturnType>(src,n,i);
    }
    template<int N>
    static Array<N+1,ReturnType>
    Extend(const Array<N,Fix<T0> > &src, IXTYPE l0, IXTYPE u0) {
      return Array<N+1,ReturnType>().Extend(src,l0,u0);
    }
  };
#else
  /*
  template<int N, class T0>
  inline Array<N-1,Fix<T0> >
  Rmdim(const Array<N,Fix<T0> > &src, IXTYPE i) {
    Array<N-1,Fix<T0> > ret;
    Ix<N-1> len0,l0;
    for(int n=0; n<N-1; n++)
      len0[n]=src.Len(n);
    for(int n=0; n<N-1; n++)
      l0[n]=src.L(n);

    ret.Realloc(len0,l0,src.Data(),src.Origo());
    return ret;
  }
  */
  template<int N, class T0>
  inline Array<N-1,typename Array<N,Fix<T0> >::T>
  Rmdim(const Array<N,Fix<T0> > &src, DIMT n, IXTYPE i) {
    return Array<N-1,typename Array<N,Fix<T0> >::T>().Rmdim(src,n,i);
  }
  template<int N, class T0>
  inline Array<N+1,typename Array<N,Fix<T0> >::T>
  Extend(const Array<N,Fix<T0> > &src, IXTYPE l0, IXTYPE u0) {
    return Array<N+1,typename Array<N,Fix<T0> >::T>().Extend(src,l0,u0);
  }
#endif

}

#endif
