/* -*- c++ -*- */
/*
 * Expresso, a C++ Array template class library
 * Copyright (C) 1998,2004 Oskar Enoksson (enok@lysator.liu.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#ifndef EXPRESSO_IX_HDR
#define EXPRESSO_IX_HDR

#include "debug.h"
#include "tmpltool.h"
#include "ixtype.h"

namespace esso {

  template<int N, class T=IXTYPE>
  class Ix;

  namespace IndexInternal {

    template<int K, int M, bool Method=(M>K)>
    struct Recurse;

    template<int K, int M>
    struct Recurse<K,M,true> {
      template<int N, class T>
      inline static T Prod(const Ix<N,T> &d) { 
	return d[M-1]*Recurse<K,M-1,(M-1>K)>::Prod(d); }
      template<int N, class T>
      inline static T SumProd(const Ix<N,T> &s1, const Ix<N,T> &s2) { 
	return s1[M-1]*s2[M-1]+Recurse<K,M-1,(M-1>K)>::SumProd(s1,s2); }
    };

    template<int K, int M>
    struct Recurse<K,M,false> {
      template<int N, class T>
      inline static T Prod(const Ix<N,T> &d) { 
        return 1; }
      template<int N, class T>
      inline static T SumProd(const Ix<N,T> &s1, const Ix<N,T> &s2) { 
	return 0; }
    };

    template<class T0>
    struct Info {
      typedef T0 T;
      enum { len=1 };
      enum { totlen=1 };
      enum { rank=0 };
    };

    template<class T0, int N>
    struct Info<Ix<N,T0> > {
      typedef typename Info<T0>::T T;
      enum { len=N };
      enum { totlen=len*Info<T0>::totlen };
      enum { rank=1+Info<T0>::rank };
    };
  }

  /// Represents a small, fast vector
  /** 
   * Use this object for small vectors where the length of the 
   * vector is known at compile time.
   * (It is possible to use it for larger datasets too, but it should
   * be noted that all data is allocated on the stack, no dynamic memory
   * is allocated by these objects).
   */
  template<int N, class T0>
  class Ix {
  public:
    typedef typename IndexInternal::Info<T0>::T T;
    template<int NN, class TT>
    friend class Ix;
    ///@name construction
    //@{
    /** Default constructor. Just creates the vector without initializing
     * the elements */
    inline Ix() {}
    /** Constructor which inializes all elements with scalar (copied 
     * into all elements). */
    explicit inline Ix(T i0) {
      for(DIMT n=0; n<N; n++) d[n]=i0;
    }
    /** Copy constructor. Copies elementwise */
    inline Ix(const Ix &i0) {
      for(DIMT n=0; n<N; n++) d[n]=i0[n];
    }
    //@}
    ///@name construction from sequence of elements
    //@{
    /** Constructor from two integers (only compiles for Ix<2,T>) */
    inline Ix(T0 i0, T0 i1) {
      TmplTools::TAssert<(N==2)>();
      d[0]=i0, d[1]=i1;
    }
    /** Constructor from three integers (only compiles for Ix<3,T>) */
    inline Ix(T0 i0, T0 i1, T0 i2) {
      TmplTools::TAssert<(N==3)>();
      d[0]=i0, d[1]=i1, d[2]=i2;
    }  
    /** Constructor from four integers (only compiles for Ix<4,T>) */
    inline Ix(T0 i0, T0 i1, T0 i2, T0 i3) {
      TmplTools::TAssert<(N==4)>();
      d[0]=i0, d[1]=i1, d[2]=i2, d[3]=i3;
    }
    /** Constructor from five integers (only compiles for Ix<5,T>) */
    inline Ix(T0 i0, T0 i1, T0 i2, T0 i3, T0 i4) {
      TmplTools::TAssert<(N==5)>();
      d[0]=i0, d[1]=i1, d[2]=i2, d[3]=i3, d[4]=i4;
    }  
    /** Constructor from six integers (only compiles for Ix<6,T>) */
    inline Ix(T0 i0, T0 i1, T0 i2, T0 i3, T0 i4, T0 i5) {
      TmplTools::TAssert<(N==6)>();
      d[0]=i0, d[1]=i1, d[2]=i2, d[3]=i3, d[4]=i4, d[5]=i5;
    }
    //@}
    ///@name const
    //@{
    /** Elementwise access */
    inline T0 operator [](DIMT n) const { 
      EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N); 
      return d[n]; 
    }
    /**@return true if this object is elementwise equal to \a src */
    inline bool operator ==(const T &src) const {
      for(DIMT n=0; n<N; n++)
	if(!(d[n]==src))
	  return false;
      return true;
    }
    /**@return true if this object is elementwise equal to \a src */
    inline bool operator ==(const Ix &src) const {
      for(DIMT n=0; n<N; n++)
	if(!(d[n]==src[n]))
	  return false;
      return true;
    }
    /**@return true if this object is not elementwise equal to \a src */
    inline bool operator !=(const T &src) const {
      for(DIMT n=0; n<N; n++)
	if(d[n]!=src)
	  return true;
      return false;
    }
    /**@return true if this object is not elementwise equal to \a src */
    inline bool operator !=(const Ix &src) const {
      for(DIMT n=0; n<N; n++)
	if(d[n]!=src[n])
	  return true;
      return false;
    }
    /**@return an \c Ix object whose elements are a sub-range
     * of the elements of this \c Ix object */
    template<DIMT LEN>
    inline Ix<LEN,T0> Sub(DIMT start=0) const {
      TmplTools::TAssert<(LEN<=N)>();
      EXPRESSO_DEBUG_CHECK_DIM(0<=start && LEN+start<=N); 
      Ix<LEN,T0> tmp;
      for(int n=0; n<LEN; n++)
	tmp[n] = d[start+n];
      return tmp;
    }
    //@}
    ///@name nonconst
    //@{
    /** Elementwise access */
    inline T0 & operator [](DIMT n) { 
      EXPRESSO_DEBUG_CHECK_DIM(0<=n && n<N); 
      return d[n]; 
    }
    /** Assignment from scalar (value is copied into all elements) */
    inline Ix & operator =(T src) {
      for(DIMT n=0; n<N; n++) d[n]=src;
      return *this;
    }
    /** Addition */
    template<class TT>
    inline Ix & operator +=(const Ix<N,TT> &src) {
      for(DIMT n=0; n<N; n++) d[n]+=src.d[n];
      return *this;
    }
    /** Addition with scalar */
    inline Ix & operator +=(const T &src) {
      for(DIMT n=0; n<N; n++) d[n]+=src;
      return *this;
    }
    /** Subtraction */
    template<class TT>
    inline Ix & operator -=(const Ix<N,TT> &src) {
      for(DIMT n=0; n<N; n++) d[n]-=src.d[n];
      return *this;
    }
    /** Subtraction with scalar */
    inline Ix & operator -=(const T &src) {
      for(DIMT n=0; n<N; n++) d[n]-=src;
      return *this;
    }
    /** Multiplication with scalar */
    inline Ix & operator *=(const T &src) {
      for(DIMT n=0; n<N; n++) d[n]*=src;
      return *this;
    }
    /** Quotient with scalar */
    inline Ix & operator /=(const T &src) {
      for(DIMT n=0; n<N; n++) d[n]/=src;
      return *this;
    }
    /** Concatenate two vectors i.e. 
     * if v1 has dim n1 and v2 has dim n2 then
     * Concat(v1,v2)[n] == v1[n] if n<n1 and v2[n-n1] otherwise
     *@return reference to this object
     */
    template<int N1>
    inline Ix &Concat(const Ix<N1,T0> &p1, 
		      const Ix<N-N1,T0> &p2) {
      for(DIMT n=0; n<N1; n++)
	d[n] = p1[n];
      for(DIMT n=0; n<N-N1; n++)
	d[n+N1] = p2[n];
      return *this;
    }
    /**@return Concatenation of this two Ix objects
     *@see Ix::Concat
     */
    template<int M>
    inline Ix<N+M,T0> operator |(const Ix<M,T0> &p2) const {
      return Ix<N+M,T0>().Concat(*this,p2);
    }
    //@}
  protected:
    /**\internal Actual element data array */
    T d[N];
  };

  ///@name Ix functions
  //@{
  /** Unary negation */
  template<int N,class T> inline Ix<N,T> operator - (const Ix<N,T> &ix) { 
    Ix<N,T> ret; 
    for(DIMT n=0; n<N; n++) 
      ret[n]=-ix[n];
    return ret; 
  } 
  /** Addition, object + scalar */
  template<int N,class T> inline Ix<N,T> operator + (const Ix<N,T> &ix, T i) { 
    Ix<N,T> ret; 
    for(DIMT n=0; n<N; n++) 
      ret[n]=ix[n] + i; 
    return ret; 
  } 
  /** Addition */
  template<int N,class T> inline Ix<N,T> operator +(const Ix<N,T> &ix, const Ix<N,T> &ixx) { 
    Ix<N,T> ret; 
    for(DIMT n=0; n<N; n++) 
      ret[n]=ix[n] + ixx[n]; 
    return ret; 
  } 
  /** Addition scalar + object */
  template<int N,class T> inline Ix<N,T> operator + (T i, const Ix<N,T> &ix) { 
    Ix<N,T> ret; 
    for(DIMT n=0; n<N; n++) 
      ret[n]=i + ix[n]; 
    return ret; 
  }
  /** Subtraction object - scalar */
  template<int N,class T> inline Ix<N,T> operator - (const Ix<N,T> &ix, T i) { 
    Ix<N,T> ret; 
    for(DIMT n=0; n<N; n++) 
      ret[n]=ix[n] - i; 
    return ret; 
  } 
  /** Subtraction */
  template<int N,class T> inline Ix<N,T> operator -(const Ix<N,T> &ix, const Ix<N,T> &ixx) { 
    Ix<N,T> ret; 
    for(DIMT n=0; n<N; n++) 
      ret[n]=ix[n] - ixx[n]; 
    return ret; 
  } 
  /** Subtraction scalar - object */
  template<int N,class T> inline Ix<N,T> operator - (T i, const Ix<N,T> &ix) { 
    Ix<N,T> ret; 
    for(DIMT n=0; n<N; n++) 
      ret[n]=i - ix[n]; 
    return ret; 
  }
  /** Product object * scalar */
  template<int N,class T> inline Ix<N,T> operator * (const Ix<N,T> &ix, T i) { 
    Ix<N,T> ret; 
    for(DIMT n=0; n<N; n++) 
      ret[n]=ix[n] * i; 
    return ret; 
  } 
  /** Product scalar * object */
  template<int N,class T> inline Ix<N,T> operator * (T i, const Ix<N,T> &ix) { 
    Ix<N,T> ret; 
    for(DIMT n=0; n<N; n++) 
      ret[n]=i * ix[n]; 
    return ret; 
  }
  /** Quotient object / scalar */
  template<int N,class T> inline Ix<N,T> operator / (const Ix<N,T> &ix, T i) { 
    Ix<N,T> ret; 
    for(DIMT n=0; n<N; n++) 
      ret[n]=ix[n] / i; 
    return ret; 
  } 
  /** Quotient scalar / object */
  template<int N,class T> inline Ix<N,T> operator / (T i, const Ix<N,T> &ix) { 
    Ix<N,T> ret; 
    for(DIMT n=0; n<N; n++) 
      ret[n]=i / ix[n]; 
    return ret; 
  }
  //@}
}

#endif
